(* This version USCI runs on noForth C2553 version 190101. Do not the use
   the low power version, because this code jumps in the new MS-routine ;)

  USCI hardware SPI on MSP430G2553 using port-1 & port-2.
  SPI i/o interfacing the nRF24L01 with two or more Launchpad boards
  Micro Launchpads and/or Egel kits.

  Connect the SPI lines of USCIB P1.5=CLOCKPULSE, P1.6=DATA-IN, P1.7=DATA-OUT
  P1.4=CSN,  P2.3=CE of the nRF24L01. On the Egel kit it's just putting the
  module in the connector marked nRF24L01!!

  Note that decoupling is very important right near the nRF24l01 module. The
  Egel kit vsn-2 has an extra 22uF near the power connections. The Launchpad
  and the Egel kit vsn-1 need an extra 10uF or more for decoupling!!

  More info on page 445 of SLAU144J.PDF  Configuration of the pins on
  page 49 of SLAS735J.PDF

  The user words are: CHECK CARRIER PULSE & WAVE

  SPI USCI master

                     MSP430G2xx3
               ^  -----------------
              /|\|              XIN|- Optional 32.768 kHz xtal
               | |                 |
               --|RST          XOUT|- Idem
                 |                 |
           IRQ ->|P2.5         P1.7|-> Data Out (UCB0SIMO)
                 |                 |
            CE <-|P2.3         P1.6|<- Data In (UCB0SOMI)
                 |                 |
           CSN <-|P1.4         P1.5|-> Serial Clock Out (UCB0CLK)
                 |                 |
                 |             P2.4|-> Power out

  Concept: Willem Ouwerkerk & Jan van Kleef, october 2014
  Current version: Willem Ouwerkerk, januari 2019

  Launchpad documentation for USCI SPI

P1 & P2 are used for interfacing the nRF24L01+
P2.3  - CE                      \ Device enable high  x1=Enable
P2.5  - IRQ                     \ Active low output   x0=Interrupt
P1.4  - CSN                     \ SPI enable low      x1=Select
P1.5  - CLOCKPULSE              \ Clock               x1=Clock
P1.6  - DATA-IN                 \ Data bitstream in   x0=Miso
P1.7  - DATA-OUT                \ Data bitstream out  x1=Mosi
P2.0  - Led red                 \ XEMIT led
P2.4  - Power led               \ Power output 2 Amp.

P1.0  - Analog input            \ 0 to 3V3
P1.3  - Switch S2

Free I/O:
P2.1  -
P2.2  -
P2.6  -
P2.7  -


Used register adresses:
020 = P1IN      - Input register
021 = P1OUT     - Output register
022 = P1DIR     - Direction register
026 = P1SEL     - Configuration register 1
027 = P1REN     - Resistance on/off
041 = P1SEL2    - Configuration register 2

029 = P2OUT     - Output register
02A = P2DIR     - Direction register
02E = P2SEL     - Configuration register 1
02F = P2REN     - Resistance on/off

nRF basic timing:
    Max. SPI clock:         8 MHz
    Powerdown to standby:   1,5 ms
    Standby to TX/RX:       0,13 ms
    Transmit pulse CE high: 0,01 ms

Sensitivity: 2Mbps  -83dB, 1Mbps  87dB, 250kbps  -96dB.
Receiving current: 2Mbps  15mA, 1Mbps  14.5mA, 250kbps  14mA

On board PCB antenna, transmission distance reach 240M in open area, but
2.4G frequency is not good to pass through walls, also interfere by
2.4G wifi signal significantly.

Record format thirteen bytes:
Node 0:  |Admin|Dest.|Orig.|Sub node|Command|D0| to |D7|  Command
Node X:  |Admin|Dest.|Orig.|Sub node|Command|D0| to |D7|  (Re)Transmit
Node Y:  |Admin|Dest.|Orig.|Sub node|Command|D0| to |D7|  Answer

00 pay>       = Administration byte
01 pay>       = Destination node
02 pay>       = Origin node
03 pay>       = Address of sub node at destination
04 pay>       = Command for destination node
05 to 11 pay> = Data bytes (0C for now)

*)

hex
value T?                \ Tracer on/off
: TEMIT     t? if  dup emit  then  drop ;  \ Show copy of char
: T"        postpone t?  postpone if  postpone ."  postpone then ; immediate
: T.        t? if  dup u.  then  drop ;
: TCR       t? if  cr  then ;
: TRON      true to t? ;    : TROFF     false to t? ;

: RED-ON    01 29 *bis ;    : RED-OFF   01 29 *bic ;
code /MS    ( u -- )  4039 , C7 , 4030 , ' ms 6 + ,  end-code   \ 06-10-2018

code IRQ?   ( -- flag ) \ Flag is true when IRQ = low
    8324 , 4784 , 0 ,   \ tos sp -) mov
    B0F2 , 20 ,  28 ,   \ 20 # 28 & .b bit \ P2IN
    7707 ,              \ tos tos subc
    4F00 ,              \ next
end-code


                    ( USCI-B SPI interface to nRF24L01 )

: SPI-SETUP     ( -- )
    01 069 *bis     \ UCB0CTL1  Reset USCI
    08 021 c!       \ P1OUT     Pullup on P1.3
    F0 022 c!       \ P1DIR     P1.4, P1.5, P1.6 and P1.7 output
    08 27 *bis      \ P1REN     P1.3  Resistor on
    E0 026 *bis     \ P1SEL     P1.5 P1.6 P1.7 is SPI
    E0 041 *bis     \ P1SEL2
    19 2A c!        \ P2DIR     P2.3 & P2.5 for SPI, P2.0=Led, P2.4 is power out
    20 2F *bis      \ P2REN     P2.5 resistor on
    21 29 *bis      \ P2OUT     P2.5 pullup, led on
    A9 068 *bis     \ UCB0CTL0  Clk=low, MSB first, Master, Synchroon
    80 069 *bis     \ UCB0CTL1  USCI clock = SMClk
    04 06A c!       \ UCB0BR0   Clock is 8Mhz/4 = 2MHz
    00 06B c!       \ UCB0BR1
    00 06C c!       \ UCB0MCTL  Not used must be zero!
    01 069 *bic     \ UCB0CTL1  Free USCI
    09 29 *bic      \ P2OUT  ce  = 0, Red led off
    10 21 *bis      \ P1OUT  csn = 1
    20 21 *bic ;    \ P1OUT  clk = 0

code SPI-I/O    ( b1 -- b2 )            \ Read and write at SPI-bus
    B2F2 , 03 ,         \ begin,  #8 03 & .b bit \ IFG2
    2BFD ,              \ cs? until,
    47C2 , 6F ,         \ tos  6F & .b mov \ UCB0TXBUF
    B2E2 , 03 ,         \ begin,  #4 03 & .b bit \ IFG2
    2BFD ,              \ cs? until,
    4257 , 6E ,         \ 6E &  tos .b mov \ UCB0RXBUF
    4F00 ,              \ next
end-code

: SPI-OUT   ( b -- )    spi-i/o drop ;  \ Write x to SPI-bus
: SPI-IN    ( -- b )    0 spi-i/o ;     \ Read b from SPI-bus


: ACTIVATE     ( -- )      10 21 *bic ; \ P1OUT  SPI on CSN=low
: DEACTIVATE   ( -- )      10 21 *bis ; \ P1OUT  SPI off CSN=high
: TRANSMIT     ( -- )      08 29 *bis noop noop 08 29 *bic ; \ P2OUT  Transmit pulse on CE


                ( Read and write to and from nRF24L01 )

55 constant #CH     \ Used channel number
100 constant #IRQ   \ Delay loops for -IRQ?
value DEST          \ Destination address
value DELAY         \ Delay for READ? the receive wait loop
value TP            \ Remember ping delay time
: RMASK    ( b1 -- b2 )    1F and ;
: WMASK    ( b1 -- b2 )    1F and  20 or ;

\ The first written byte returns internal status always
\ It is saved in the value STATUS using SPI-COMMAND
value STATUS
: GET-STATUS    ( -- s )    activate  FF spi-i/o  deactivate ;
: SPI-COMMAND   ( c -- )    activate  spi-i/o to status ;

\ Reading and writing to registers in the 24L01
: READ-REG      ( r -- b )  rmask spi-command  spi-in  deactivate ;
: WRITE-REG     ( b r -- )  wmask spi-command  spi-out deactivate ;

\ Read and write the communication addresses, of pipe-0 default: E7E7E7E7E7
\ : READ-ADDR    ( trxa -- ) rmask spi-command  5 0 do spi-in  loop  deactivate ;
: WRITE-ADDR  ( trxa -- )  wmask spi-command  5 0 do spi-out loop  deactivate ;


                ( nRF24L01 control commands and setup )

\ Empty RX or TX data pipe
: FLUSH-RX      ( -- )      E2 spi-command deactivate ; \ Remove received data
: FLUSH-TX      ( -- )      E1 spi-command deactivate ; \ Remove transmitted data
: RESET         ( -- )      70 07 write-reg ;           \ Reset IRQ flags
\ : PIPE          ( -- +n )   07 read-reg 0E and ;        \ Leave filled RX-pipe nr.
: PIPES-ON      ( mask -- ) 3F and  02 write-reg ;      \ Set active pipes
: WAKEUP        ( -- )      0E 00 write-reg ; \ CRC 2 bytes, Powerup
: >CHANNEL      ( +n -- )   05 write-reg ;    \ Change RF-channel 7-bits
: >NRF          ( +n -- )   06 write-reg ;    \ 26 = 250kbps, full power
                                              \ 06 = 1 Mbit, full power (0 dBm)
                                              \ 00 = 1 Mbit, -18 dBm = lowest power
                                              \ 0E - 2 Mbit, full power
                                              \ 0C = 2 Mbit, -6 dBm = lower power


\ Elementary command set for the nRF24L01
value NRF               \ Contains choosen nRF24 RF setup
: DEFAULT-RF    ( -- )  06 to nrf ; \ Set default RF mode
default-rf

11 constant #PAY        \ Payload size 17 bytes
: SETUP24L01    ( -- )
\   spi-setup
\   -1 to status        \ Init. status
\   dm 100 to delay     \ Wait 11 ms for an answer after a command
    03 01 write-reg     \ Auto Ack pipe 0 & 1
    02 pipes-on         \ Pipe 1 on
    F0 F0 F0 F0 #me 0B write-addr \ Set ME own address
    03 03 write-reg     \ Five byte address
    1F 04 write-reg     \ Retry after 500 us & 15 retry's
    #ch >channel        \ channel #CH to start with
    nrf >nrf            \ 1 Mbps, max. power
    #pay 11 write-reg   \ #pay bytes payload in P0
    #pay 12 write-reg   \ #pay bytes payload in P1
    flush-tx  flush-rx  \ Start empty
    wakeup 0F /ms reset \ Enable CRC, 2 bytes, power up & reset flags
    red-off ;           \ Led off


\ Format: Admin, Dest. node, Org. node, Sub node, Command, Data-0, Data-1, Data-2, Data-3
create 'READ    #pay allot  \ Receive buffer
create 'WRITE   #pay allot  \ Transmit buffer
\ : .BUFF     ( a -- )    #pay bounds ?do  i c@ .  loop ;  \ Show buff.

: READ-RX   ( -- )      \ Receive #PAY bytes
    61 spi-command  'read #pay bounds
    ?do  spi-in i c!  loop   deactivate ;

: WRITE-TXA ( -- )      \ Send #PAY bytes with Ack check
    A0 spi-command  'write #pay bounds
    ?do  i c@ spi-out  loop  deactivate ;

: '>PAY     ( +n -- a )     'write + ;  \ Leave address of TX payload
: >PAY      ( b +n -- )     '>pay c! ;  \ Store byte for TX payload
: 'PAY>     ( +n -- a )     'read + ;   \ Leave address of RX payload
: PAY>      ( +n -- b )     'pay> c@ ;  \ Read byte from RX payload
: THRU      ( -- )      'read 'write #pay move ; \ Copy RX- to TX-payload

: .STATUS   ( -- )
    ." Node " #me .  ." nRF24 " \ Which node with nRf24
    get-status 0E <> if         \ Not well?
      ." not ok "  abort        \ Yes, show & stop
    then  ."  ok " ;            \ All ok, go on

\ Different tests for receiver and transmitter status
: RX?           ( -- f )    07 read-reg 40 and 0<> ;    \ Data received
: TX?           ( -- f )    07 read-reg 20 and 0<> ;    \ Ack received
\ : FIFO-EMPTY?   ( -- f )    17 read-reg 01 and 0<> ;    \ Pipeline empty
\ : FIFO-FULL?    ( -- f )    17 read-reg 01 and 0= ;     \ Pipeline full
\ : MAXRETRY?     ( -- f )    07 read-reg 10 and 0<> ;    \ Transmit failed
\ : PIPE?         ( +n -- f ) 2* pipe = ;                 \ Correct Pipe?
\ : >DYNPD        ( +n -- )   1C write-reg ;  \ enable dynamic payload
\ : >FEATURE      ( +n -- )   1D write-reg ;  \ Payload properties


                    ( Send and receive commands for nRF24L01 )

: WRITE-MODE    ( -- )          \ Power up module as transmitter
    08 29 *bic  wakeup          \ P2OUT  CE low, receive off, wakeup transmitter
    reset  1 pipes-on  2 /ms ;  \ Reset flags & pipe-0 active, wait 200 microsec.

: READ-MODE     ( -- )
    0F 00 write-reg  2 pipes-on \ Power up module as receiver, activate pipe-1
    reset  08 29 *bis  2 /ms ;  \ P2OUT  Enable receive mode, wait 200 microsec.

: -IRQ?         ( -- )      \ Wait for IRQ
    #irq 0 ?do              \ See top of file, used to be: 200
        irq? if             \ IRQ happened?
            unloop  exit    \ Yes, ready
        then
    loop ;

: XEMIT)        ( c -- )    \ Send #PAY bytes payload ARC times
    4 >pay  write-txa  transmit  -irq? ; \ Send command

: XEMIT?        ( c -- +n ) \ +n = 0A when transmit has failed
    0  0A 0 do              \ Try it 0A times * ARC = dm 150 times max.
        over xemit)         \ Send payload
    tx? if  leave  then     \ Ready when it was an Ack!
        reset  1+           \ Reset & count failures
        flush-tx  #ch >channel \ Empty pipeline, clear packet loss
    loop
    nip  flush-tx  reset ;


: READ?         ( -- fl )   \ Wait for receive IRQ
    delay 0 ?do  1 /ms  irq? if  i to tp  true unloop exit  then  loop  false ;

: XKEY          ( -- c )
    begin
        begin  irq? until \ Something happened?
    rx? until           \ Is it a payload
    read-rx  4 pay>     \ Read command from packet
    reset  flush-rx     \ Empty pipeline
    08 29 *bic ;        \ To standby II

\ Set destination address to node from stack, receive address is my "me"
: SET-DEST      ( node -- )
    dup to dest  1 >pay  #me 2 >pay \ Set Destination & origin nodes
    F0 F0 F0 F0 dest 0A write-addr  \ Receive address P0
    F0 F0 F0 F0 dest 10 write-addr ; \ Transmit address

\ Send command & receive data while doing maximal x 'retries'
\ The result flag is false when all is ok.
: TRANSCEIVE    ( cT retries -- flag )
    0 ?do
        red-on  write-mode      \ Led on, go transmitting
        dup xemit?  0A < if     \ Send char command
            reset  read-mode    \ Reset flags, go receiving
            read? if            \ Data received?
                drop  xkey drop \ Read payload, success
                08 029 *bic  false \ P2OUT  CE low, receive mode off
                red-off  unloop exit \ Ready!
            then
        then
    loop
    drop  red-off  08 29 *bic   \ P2OUT  CE low, receive mode off
    true ;                      \ Transceive failed somewhere!


shield 24L01\   freeze

\ End
